/*
 * Copyright (C) 2004 USAGI/WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include <stdlib.h>
#include <stdio.h>
#include <netinet/in.h>
#include <netinet/ip6.h>
#include <netinet/icmp6.h>
#include <netinet/udp.h>

#define DATALEN 65527
#define NUM_FRAGS 4

unsigned int
csum_fold(unsigned int sum)
{
	/* TODO: this depends on byte order */

	while (sum >> 16)
		sum = (sum & 0xffff) + (sum >> 16);

	return (u_int16_t) ~sum;
}

unsigned int
csum_no_fold(u_int16_t *packet, int packlen, unsigned int sum)
{
	register unsigned int new = sum;

	while (packlen > 1) {
		new+= *(packet++);
		packlen-=2;
	}

	if (packlen > 0)
		new += *(unsigned char *)packet;

	return new;
}

u_int16_t
csum(u_int16_t *packet, int packlen)
{
	register unsigned int sum = 0;

	while (packlen > 1) {
		sum+= *(packet++);
		packlen-=2;
	}

	if (packlen > 0)
		sum += *(unsigned char *)packet;

	return csum_fold(sum);
}

u_int16_t
csum_ipv6_magic(struct in6_addr *saddr, struct in6_addr *daddr, u_int32_t len,
		unsigned short proto, unsigned int sum)
{
	int i;

	for (i=0; i<8; i++)
		sum += saddr->s6_addr16[i];
	for (i=0; i<8; i++)
		sum += daddr->s6_addr16[i];
	sum += htonl(len);
	sum += htons(proto);

	return csum_fold(sum);
}

void
usage(char *name) {
	fprintf(stderr, "usage: %s src_addr src_prt dst_addr dst_prt\n", name);
}

int
main(int argc, char *argv[])
{
	struct in6_addr src_addr, dst_addr;
	u_int16_t src_prt, dst_prt;
	u_int16_t datalen = DATALEN;
	u_int16_t udp_size;
	struct udphdr udph;
	unsigned int check = 0;
	char *frags[NUM_FRAGS];
	int frag_size[NUM_FRAGS] = {256, 256, 264, 640};
	int frag_offset[NUM_FRAGS] = {0, 512, 768, 256};
	int frag_mf[NUM_FRAGS] = {1, 1, 1, 1};
	char *data;
	int i;
	struct ip6_frag *fragh_ptr;
	char *data_ptr;

	int s;
	struct sockaddr_in6 sin;
	int ret;
	int id;

	if (argc != 5) {
		usage(argv[0]);
		exit(1);
	}

	if (inet_pton(AF_INET6, argv[1], (struct in6_addr *)&src_addr) <= 0) {
		fprintf(stderr, "bad src IPv6 address: %s\n", argv[1]);
		usage(argv[0]);
		exit(2);
	}

	src_prt = atoi(argv[2]);

	if (inet_pton(AF_INET6, argv[3], (struct in6_addr *)&dst_addr) <= 0) {
		fprintf(stderr, "bad dst IPv6 address: %s\n", argv[1]);
		usage(argv[0]);
		exit(3);
	}

	dst_prt = atoi(argv[4]);

	/* make data */
	data = (char *)malloc(datalen);
	memset(data, 1, datalen);

	/* make UDP header */
	memset(&udph, 0, sizeof(udph));
	udp_size = sizeof(struct udphdr) + datalen;
	udph.source = ntohs(src_prt);
	udph.dest = ntohs(dst_prt);
	udph.len = ntohs(udp_size);
	check = csum_no_fold((u_int16_t *)data, datalen, 0);
	check = csum_no_fold((u_int16_t *)&udph, sizeof(udph), check);
	udph.check = csum_ipv6_magic(&src_addr, &dst_addr, udp_size,
				     IPPROTO_UDP, check);

	/* make fragmented packets */

	srand(getpid());
	id = rand();

	for (i=0; i<NUM_FRAGS; i++) {

		frags[i] = (char *)malloc(sizeof(struct ip6_frag)
						     + frag_size[i]);

		fragh_ptr = (struct ip6_frag *)frags[i];
		data_ptr = (char *)fragh_ptr + sizeof(struct ip6_frag);

		fragh_ptr->ip6f_nxt = IPPROTO_UDP;
		fragh_ptr->ip6f_reserved = 0;
		fragh_ptr->ip6f_offlg = htons(frag_offset[i]);
		fragh_ptr->ip6f_ident = id;

		if (frag_offset[i] < sizeof(udph)) {
			if (frag_offset[i] + frag_size[i] < sizeof(udph)) {
				memcpy(data_ptr, (char *)&udph + frag_offset[i],
				       frag_size[i]);
			} else {
				memcpy(data_ptr, (char *)&udph + frag_offset[i],
				       sizeof(udph) - frag_offset[i]);

				memcpy(data_ptr + sizeof(udph) - frag_offset[i],
				       data,
				       frag_size[i] - sizeof(udph)
				       + frag_offset[i]);
			}
		} else {
			memcpy(data_ptr, data + frag_offset[i] - sizeof(udph),
			       frag_size[i]);
		}

		if (frag_mf[i])
			fragh_ptr->ip6f_offlg |= IP6F_MORE_FRAG;
	}


	/* prepare to send msg */
	s = socket(PF_INET6, SOCK_RAW, IPPROTO_FRAGMENT);

	if (s < 0) {
		perror("socket: can't open socket");
		exit(4);
	}

	memset(&sin, 0, sizeof(sin));
	sin.sin6_family = AF_INET6;

	memcpy(sin.sin6_addr.s6_addr, dst_addr.s6_addr, sizeof(dst_addr));

	for (i=0; i<NUM_FRAGS; i++) {
		ret = sendto(s, (void *)frags[i],
			     sizeof(struct ip6_frag) + frag_size[i], 0,
			     (struct sockaddr *)&sin, sizeof(sin));
		if (ret < 0) {
			perror("sendto: failed to send");
		}
	}
}
